/*
****************************************************************************

            Copyright 1991, by FUJITSU LIMITED.
            Copyright 1991, by Sun Microsystems, Inc.

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Sun Microsystems, Inc. or
FUJITSU LIMITED not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.  FUJITSU LIMITED and Sun Microsystems, Inc. make no
representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

SUN MICROSYSTEMS, INC. AND FUJITSU LIMITED DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL FUJITSU LIMITED AND SUN
MICROSYSTEMS, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CTrueSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTITrue OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTITrue, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

  Author: A.Yukishita, FUJITSU LIMITED
          Riki Kawaguchi FUJITSU LIMITED

****************************************************************************
*/

#ifndef lint
#ident	"@(#)HttLkup.c	1.50	93/03/09 SMI"
#endif

#ifdef X_LOCALE
#include <X11/Xlocale.h>
#else
#include <locale.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#if XlibSpecificationRelease != 5
#include <X11/XlibR5.h>
#endif /* XlibSpecificationRelease */
#include "XSunExt.h"

#include <X11/Xutil.h>
#include <X11/Xresource.h>

#include <X11/keysym.h>
#include <X11/Xatom.h>

#include "Htt.h"
#include "XSunIMProt.h"
#include "HttXsfSun.h"
#include "HttLkup.h"

#define LKUP_X			50
#define LKUP_Y			100
#define	LKUP_POPUP_SPACE	10
#define	LKUP_NEXT		"^N:NEXT"
#define	LKUP_NEXT_LEN		7
#define	LKUP_PRIOR		"^P:PRIOR"
#define	LKUP_PRIOR_LEN		8

static void	Lkup_XIMIsMaster() ;
static void	Lkup_CBIsMaster() ;
static void	Lkup_cancel_item();
static void	Lkup_make_item();
static int 	Lkup_get_item();
static void	Lkup_resize_window();
static void	Lkup_draw_line();
static void	Lkup_draw_item();
static int	xsf_wcscmp();
static XFontSet	Lkup_get_font();
static void	Lkup_ConstructNegoData();
static void	CreatePopupLucWindows();
static void	CreateRootLucWindow();
static void	SetupPopupLucGeometry();
static void	SetupRootLucGeometry();
static void	SwitchToPopupStyle();
static void	SwitchToRootStyle();
static Bool	WMPopupLucWindows();
static void	MapIMSLuc();
static void     GetLayout();

extern unsigned long 	GetPixelByName();
extern void	Lkup_Redisplay();
extern void	Lkup_End();
extern void     stipple_or_unmap_preedit_window();
extern void     stipple_or_unmap_status_window();
extern void     MapStatusWindow();
extern void     draw_status_area();


static  Window           ims_map_luc = -1;
static  FXimp_Client *current_lookup_ic = 0;

static wchar_t		constant_alph[64];

typedef	struct _LookupItem {
    Bool            page_start_flag;
    wchar_t        *label;
    wchar_t        *string;
    XIMFeedback	   *feedback;		/* Feedback for string only */
    int             label_length;
    int             value_length;
    int             string_length;
    int             string_width;
    int             ncolumns;
    int             nrows;
    int             candidate;
    XRectangle      position;
} LookupItem;

typedef enum {
    LUCStart = 1,
    LUCDraw  = 2,
    LUCProc  = 3,
} LUCState;

typedef struct _LookupData {
    LUCState	    lucstate;
    XRectangle      max_position;
    XRectangle      position;
    Bool            draw_flag;
    Bool            next_page;
    Bool            prior_page;
    WhoIsMaster     who_is_master;
    LayoutInfo      layout;
    int             ncolumns;
    int             nrows;
    int             first_candidate ;
    int             last_candidate ;
    int             current_candidate;
    int             current_count;
    int             n_choices;
    int             next_page_width;
    int             prior_page_width;
    int		    title_x;
    int		    title_y;
    wchar_t	   *title;
    int		    title_len;
    XIMFeedback	   *title_fb;
    int		    title_pos;
    int		    lineoffset_x;
    int		    lineoffset_y;
    LookupItem     *lookup_item;
} LookupData ;

s_lookup_init(htt)
    htt_t			*htt;
{
    LookupInfo			*lookup_info = 
					(LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display			*display = GET_HTT_VALUE(htt, display);
    XFontSet			current_luc_font;
    char			**missing_charsets;
    int				missing_char_counts;
    char			*def_string;

    mbstowcs( lookup_info->constant_next, LKUP_NEXT, LKUP_NEXT_LEN );     
    mbstowcs( lookup_info->constant_prior, LKUP_PRIOR, LKUP_PRIOR_LEN );

    lookup_info->luc_win = (Window *)calloc((ScreenCount(display)+1) , sizeof(Window));
    lookup_info->gc = (GC *)malloc((ScreenCount(display)+1) * sizeof(GC));
    lookup_info->rgc = (GC *)malloc((ScreenCount(display)+1) * sizeof(GC));

    if( !strcmp(GET_HTT_VALUE(htt, font_name), lookup_info->font_name) )
	current_luc_font = CURRENT_CLIENT(htt)->preedit_font->font_set;
    else {
	current_luc_font = XCreateFontSet(display, lookup_info->font_name,
				&missing_charsets, &missing_char_counts,
				&def_string);
	if(!current_luc_font)
		current_luc_font = 
			CURRENT_CLIENT(htt)->preedit_font->font_set;
    }

    lookup_info->fontset = current_luc_font;
    lookup_info->font_ext = XExtentsOfFontSet(lookup_info->fontset);

    /*
     * Lookup Window Create
     */
    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
	lookup_info->base_index = 1;
        lookup_info->num_win = ScreenCount(display);
	CreatePopupLucWindows(display, htt, lookup_info);
	SetupPopupLucGeometry(htt, lookup_info);
    }
    else {
	lookup_info->base_index = 0;
    	lookup_info->num_win = 1;
    	CreateRootLucWindow(display, htt, lookup_info);
	SetupRootLucGeometry(htt, lookup_info);
    }

    return(0);
}

static void
CreatePopupLucWindows(display, htt, lookup_info)
    Display	*display;
    htt_t	*htt;
    LookupInfo	*lookup_info;
{
    XSetWindowAttributes	attrib;
    XGCValues		gcvalues;
    int			num;
    unsigned long	background_pixel;
    unsigned long	foreground_pixel, border_pixel;
    Colormap		default_colormap;
    FXimp_Client        *root_client = GET_HTT_VALUE(htt, root_client);
    char		name[16];
    int			base = lookup_info->base_index;

    for(num = 0; num < ScreenCount(display); num++) {
	default_colormap = DefaultColormap(display, num);
	border_pixel = GetPixelByName(display, default_colormap,
 		GET_HTT_VALUE(htt, border_color), BlackPixel(display, num));
	background_pixel = GetPixelByName(display, default_colormap,
		lookup_info->background_color, WhitePixel(display, num));

	foreground_pixel = GetPixelByName(display, default_colormap, 
			lookup_info->foreground_color, BlackPixel(display, num));
 
        lookup_info->luc_win[num+base] =
        	XCreateSimpleWindow(display, RootWindow(display, num),0,0,1,1,
                        lookup_info->borderwidth, border_pixel,
                        background_pixel );
        bzero( &attrib, sizeof( XSetWindowAttributes ) );
        attrib.backing_store = NotUseful;
	if(lookup_info->move)
		attrib.override_redirect = False;
	else
            	attrib.override_redirect = True;
	if(lookup_info->save_under)
		attrib.save_under = True;
	else
		attrib.save_under = False;
        XChangeWindowAttributes( display, lookup_info->luc_win[num+base], 
                            CWBackingStore | CWOverrideRedirect | CWSaveUnder ,
			    &attrib );
	gcvalues.function = GXcopy;
	gcvalues.foreground = foreground_pixel;
	gcvalues.background = background_pixel;
	lookup_info->gc[num+base] = XCreateGC(display, lookup_info->luc_win[num+base],
				GCBackground | GCForeground | GCFunction,
				&gcvalues);
	gcvalues.foreground = background_pixel;
	gcvalues.background = foreground_pixel;
	lookup_info->rgc[num+base]= XCreateGC(display, lookup_info->luc_win[num+base],
                                GCBackground | GCForeground | GCFunction,
                                &gcvalues);
	XSelectInput(display, lookup_info->luc_win[num+base], 
			StructureNotifyMask | ExposureMask |
			KeyPressMask | KeyReleaseMask);

    }

    lookup_info->wmhits = WMPopupLucWindows(display, htt, lookup_info);
}

static Bool
WMPopupLucWindows(display, htt, lookup_info)
    Display     *display;
    htt_t       *htt;
    LookupInfo  *lookup_info;
{
    Atom         ol_del_atom_list[2];
    Atom         ol_del_atom;
    int          ol_atom_inx = 0;
    char         name[16];
    int		 num;
    FXimp_Client *root_client = GET_HTT_VALUE(htt, root_client);

    if(!lookup_info->move)
	return True;

    ol_del_atom = XInternAtom(display, "_OL_DECOR_DEL", True);
    if( ol_del_atom == None )
	return False;
    ol_del_atom_list[ol_atom_inx++] = 
    			XInternAtom(display, "_OL_DECOR_RESIZE", True);
    ol_del_atom_list[ol_atom_inx++] = 
    			XInternAtom(display, "_OL_DECOR_HEADER", True);
    for(num = 0; num < ScreenCount(display); num++)  {
	XChangeProperty(display, lookup_info->luc_win[num+1],
				ol_del_atom, XA_ATOM, 32,
				PropModeReplace, ol_del_atom_list, 
				ol_atom_inx);
       	sprintf(name, "XIMLookupMenu%d", num);
       	XmbSetWMProperties(display, lookup_info->luc_win[num+1],
		&name[0], NULL, 0, NULL, NULL, NULL, NULL);
	XSetTransientForHint(display, lookup_info->luc_win[num+1],
			root_client->window);
    }

    return True;
}

static void
CreateRootLucWindow(display, htt, lookup_info)
    Display	*display;
    htt_t	*htt;
    LookupInfo	*lookup_info;
{
    int			screen = GET_HTT_VALUE(htt, screen);
    unsigned long	background_pixel;
    unsigned long	foreground_pixel, border_pixel;
    XGCValues		gcvalues;
    Colormap		default_colormap;
    FXimp_Client        *root_client = GET_HTT_VALUE(htt, root_client);
    Ximp_PreeditPropRec	*root_preedit = root_client->preedit;
    char                *w_name = "XIMLookupMenu0";

    default_colormap = DefaultColormap(display, screen);
    border_pixel = GetPixelByName(display, default_colormap,
                 GET_HTT_VALUE(htt, border_color), BlackPixel(display, screen));
    background_pixel = GetPixelByName(display, default_colormap,
                 GET_HTT_VALUE(htt, background_color), WhitePixel(display, screen));
    foreground_pixel = GetPixelByName(display, default_colormap, 
			GET_HTT_VALUE(htt, foreground_color), 
			BlackPixel(display, screen));
 
    lookup_info->luc_win[0] = CURRENT_CLIENT(htt)->lookup_window =
	XCreateSimpleWindow( display, root_client->window,
		    root_preedit->Area.x, root_preedit->Area.y,
		    root_preedit->Area.width, root_preedit->Area.height,
		    lookup_info->borderwidth, border_pixel, background_pixel );
	
    gcvalues.function = GXcopy;
    gcvalues.foreground = foreground_pixel;
    gcvalues.background = background_pixel;
    lookup_info->gc[0] = XCreateGC(display, lookup_info->luc_win[0],
				GCBackground | GCForeground | GCFunction,
				&gcvalues);
    gcvalues.foreground = background_pixel;
    gcvalues.background = foreground_pixel;
    lookup_info->rgc[0]= XCreateGC(display, lookup_info->luc_win[0],
                                GCBackground | GCForeground | GCFunction,
                                &gcvalues);
    XSelectInput(display, lookup_info->luc_win[0], StructureNotifyMask | ExposureMask);
    XmbSetWMProperties(display, lookup_info->luc_win[0],
		&w_name[0], NULL, 0, NULL, NULL, NULL, NULL);
}

static void
SetupPopupLucGeometry(htt, lookup_info)
    htt_t	*htt;
    LookupInfo	*lookup_info;
{
    lookup_info->wspacing1 = 2;
    lookup_info->wspacing2 = 2;
    lookup_info->hspacing = 1;
    if(lookup_info->move)
    	lookup_info->borderwidth = 0;
}

static void
SetupRootLucGeometry(htt, lookup_info)
    htt_t	*htt;
    LookupInfo	*lookup_info;
{
    lookup_info->nrows = GET_HTT_VALUE(htt, ims_line);
    (!(GET_HTT_VALUE(htt, lookup_takeover)) &&
    ((GET_HTT_VALUE(htt, ims_preedit_line) < GET_HTT_VALUE(htt, ims_line)) &&
    (lookup_info->nrows -= 2)));
    ((lookup_info->nrows <= 0) && (lookup_info->nrows = 1));
    lookup_info->ncolumns = 27;
    lookup_info->wspacing1 = 0;
    lookup_info->wspacing2 = 2;
    lookup_info->hspacing = 0;
    lookup_info->linewidth = 0;
    lookup_info->borderwidth = 0;
}

static void
SwitchToPopupStyle(htt)
    htt_t	*htt;
{
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display	*display = (Display *)GET_HTT_VALUE(htt, display);

    lookup_info->num_win = ScreenCount(display);
    lookup_info->base_index = 1;
    if(!lookup_info->luc_win[1])
	CreatePopupLucWindows(display, htt, lookup_info);
    SetupPopupLucGeometry(htt, lookup_info);
    CheckRootWindow();
}

static void
SwitchToRootStyle(htt)
    htt_t	*htt;
{
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display	*display = (Display *)GET_HTT_VALUE(htt, display);
    FXimp_Client	*root_client = (FXimp_Client *)GET_HTT_VALUE(htt, root_client);

    lookup_info->num_win = 1;
    lookup_info->base_index = 0;
    if(!lookup_info->luc_win[0])
	CreateRootLucWindow(display, htt, lookup_info);
    SetupRootLucGeometry(htt, lookup_info);
    XMapWindow(display, root_client->window);
        SET_HTT_VALUE(htt, ims_window_mapped, True);

}

void
s_lookup_start( ic, rv )
    FXimp_Client	*ic ;
    iml_inst		*rv ;
{
#ifndef SVR4
    char		*malloc();
#endif
    htt_t		*htt = IMS_OF_CLIENT(ic);
    LookupInfo		*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display		*display = GET_HTT_VALUE(htt, display);

    LookupData		*lookup;
    XIMLookupStartCallbackStruct    *lss =
			    (XIMLookupStartCallbackStruct *)&rv->operand ;
    Window		root_return;
    int			x_return, y_return;
    unsigned int	width_return, height_return;
    unsigned int	border_width_return;
    unsigned int	depth_return;
    Status		status;
    int			pos_x, pos_y;
    Window		child;
    XWindowAttributes	win_attr;
    FXimp_Client        *root_client = GET_HTT_VALUE(htt, root_client);
    XRectangle		area;
    int			nlines;

    lookup_info->lock ++;

    if(lss->WhoIsMaster == HasNotBeenNegotiated) {
	Lkup_ConstructNegoData(ic, lss);
	return;
    }

    if(!lookup_info->wmhits && lookup_info->style == LOOKUP_MENU_POPUPWINDOW)
	lookup_info->wmhits = WMPopupLucWindows(display, htt, lookup_info);

    if (ic->lookup_data) 
		free (ic->lookup_data);

    ic->lookup_data = (XPointer)malloc( sizeof( LookupData ) ) ;
    lookup = (LookupData *)ic->lookup_data;
    bzero( lookup, sizeof( LookupData ) ) ;

    if(lss->WhoIsMaster == XIMIsMaster) {
	lookup->layout = *lss->XIMPreference;
    }
    else if(lss->WhoIsMaster == CBIsMaster) {
        lookup->layout = *lss->CBPreference ;
    }

    if( lookup_info->style == LOOKUP_MENU_IMSWINDOW )
	lookup->who_is_master = CBIsMaster;
    else
        lookup->who_is_master = lss->WhoIsMaster ;

    if(lookup->layout.label_type == IM_LUC_LABEL_ALPHABETIC)
        mbstowcs(constant_alph,  
			":abcdefghijklmnopqrstuvwxyz", 27 );
    else if(lookup->layout.label_type == IM_LUC_LABEL_NUMERIC)
	mbstowcs(constant_alph,  ":1234567890", 11);
    else if(lookup->layout.label_type == IM_LUC_LABEL_ALPHA_UPPER)
	mbstowcs(constant_alph,  
			":ABCDEFGHIJKLMNOPQRSTUVWXYZ", 27 );

    switch(lookup_info->title_pos) {
	case  LOOKUP_NEXT_TO_LEFT:
		break;
	case  LOOKUP_NEXT_TO_TOP:
		break;
	default:
		if(lookup->layout.nrows == 1)
			lookup->title_pos = LOOKUP_NEXT_TO_LEFT;
		else
			lookup->title_pos = LOOKUP_NEXT_TO_TOP;
    }

    switch ( lookup_info->style ) {
    case    LOOKUP_MENU_POPUPWINDOW:
	ic->lookup_window = 
		lookup_info->luc_win[ic->screen + lookup_info->base_index];
#if	notdef  /* UNTILtoolkits suppot SpotLocation */
	if(ic->set_ic_val_mask & XIMP_PRE_SPOTL_MASK) {
		XTranslateCoordinates(display, ic->focus_window,
		    RootWindow(display, ic->screen),
		    ic->preedit->SpotLocation.x, ic->preedit->SpotLocation.y,
		    &pos_x, &pos_y, &child);
		lookup->position.x = pos_x + lookup_info->off_x;
		lookup->position.y = pos_y + lookup_info->off_y;
	}
#endif
	if( lookup_info->navigator == LOOKUP_POSITION_MOUSE ) {
		lookup->position.x = lss->event->x_root
					+lookup_info->off_x;
		lookup->position.y = lss->event->y_root
					+lookup_info->off_y;
	}
	else if( lookup_info->navigator == LOOKUP_POSITION_WINDOW ) {
	  if(ic->style & XIMPreeditPosition) {
	    XTranslateCoordinates(display, ic->focus_window, 
		RootWindow(display, ic->screen), 
		ic->preedit->SpotLocation.x, ic->preedit->SpotLocation.y,
		&pos_x, &pos_y, &child);
	    lookup->position.x = pos_x + lookup_info->off_x;
	    lookup->position.y = pos_y + lookup_info->off_y;
	  }
	  else if(ic->style & XIMPreeditArea) {
	    XTranslateCoordinates(display, ic->focus_window, 
		RootWindow(display, ic->screen), 
		ic->preedit->Area.x, ic->preedit->Area.y,
		&pos_x, &pos_y, &child);
	    lookup->position.x = pos_x + lookup_info->off_x;
	    lookup->position.y = pos_y + lookup_info->off_y;
	  }
	  else if(ic->style & XIMPreeditCallbacks) {
	    XGetWindowAttributes(display, ic->focus_window, &win_attr);
	    XTranslateCoordinates(display, ic->focus_window, 
		RootWindow(display, ic->screen), 
		win_attr.x, win_attr.y, &pos_x, &pos_y, &child);
	    if(lookup_info->next_to == LOOKUP_NEXT_TO_RIGHT) {
		lookup->position.x = pos_x + win_attr.width
				+ lookup_info->off_x;
		lookup->position.y = pos_y + lookup_info->off_y;
	    }
	    else if(lookup_info->next_to == LOOKUP_NEXT_TO_BOTTOM) {
		lookup->position.x = pos_x + lookup_info->off_x;
		lookup->position.y = pos_y + win_attr.height 
				+ lookup_info->off_y;
	    }
	    else if(lookup_info->next_to == LOOKUP_NEXT_TO_TOP) {
	        lookup->position.x = pos_x + lookup_info->off_x;
	        lookup->position.y = LKUP_Y;
	    }
	    else if(lookup_info->next_to == LOOKUP_NEXT_TO_LEFT) {
	        lookup->position.x = LKUP_X;
	        lookup->position.y = pos_y + lookup_info->off_y;
	    }
	  }
	  else {	/* Root Window Need to be FIXED */
	    lookup->position.x = LKUP_X;
	    lookup->position.y = LKUP_Y;
	  }
	}
	else {	/* Don't know what this mode is */
		lookup->position.x = LKUP_X;
		lookup->position.y = LKUP_Y;
	}
	lookup->max_position.width = DisplayWidth( display, ic->screen );
	lookup->max_position.height = DisplayHeight( display, ic->screen );
	break;
    case    LOOKUP_MENU_IMSWINDOW:
	ic->lookup_window = lookup_info->luc_win[0];
	lookup->position.x = root_client->ims_area->x;
	lookup->position.y = root_client->ims_area->y;
	lookup->max_position.width = (GET_HTT_VALUE(htt, lookup_takeover)) ? 
		root_client->ims_area->width 
		: root_client->preedit->Area.width;
	lookup->max_position.height = root_client->ims_area->height;
	break;
    }
#ifdef DEBUG
	dbg_printf("    x=%d, y=%d, wid=%d, hei=%d\n",
		lookup->position.x, lookup->position.y,
		lookup->max_position.width, lookup->max_position.height );
#endif					/* DEBUG */
    lookup->lucstate = LUCStart;
    return ;
}

static void
Lkup_ConstructNegoData(ic, lss)
    FXimp_Client			*ic ;
    XIMLookupStartCallbackStruct	*lss ;
{
    htt_t	*htt = IMS_OF_CLIENT(ic);

    GetLayout(htt, lss->CBPreference);
}

static void
GetLayout(htt, layout)
    htt_t	*htt;
    LayoutInfo	*layout;
{
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    FXimp_Client *root_client= GET_HTT_VALUE(htt, root_client);
    int		cw;
    Display	*display;
    XFontSet 	fontset;
    int         rtw, rtrow;

    layout->label_type = lookup_info->label_type;
    if(lookup_info->label_type == IM_LUC_LABEL_NUMERIC)
	layout->choice_per_window = 
		MIN(10, lookup_info->nrows * lookup_info->ncolumns);
    else
	layout->choice_per_window =
		MIN(26, lookup_info->nrows * lookup_info->ncolumns);
    layout->root_width = CURRENT_CLIENT(htt)->ims_area->width;
    layout->root_lines = GET_HTT_VALUE(htt, ims_line);

    fontset = Lkup_get_font( CURRENT_CLIENT(htt) );
    cw = XmbTextEscapement(fontset, constant_alph, 1);
    layout->e_width_per_can = 
	lookup_info->font_ext->max_logical_extent.width *
	lookup_info->wspacing2 + cw * 2;
    
    layout->max_width_per_car = lookup_info->font_ext->max_logical_extent.width;
    layout->nrows = lookup_info->nrows;
    layout->ncolumns = lookup_info->ncolumns;

    if(lookup_info->style == LOOKUP_MENU_IMSWINDOW) {
	layout->luc_is_rootwindow = 1;
	rtw = root_client->ims_area->width /
		(layout->e_width_per_can + layout->max_width_per_car);
	rtrow = GET_HTT_VALUE(htt, ims_line);
        (!(GET_HTT_VALUE(htt, lookup_takeover)) &&
	((GET_HTT_VALUE(htt, ims_preedit_line) < GET_HTT_VALUE(htt, ims_line))
	&& (rtrow -= 2)));
	((rtrow <= 0) && (rtrow = 1));
	layout->choice_per_window = 
		MIN(layout->choice_per_window, rtw * rtrow);
    }
    else
	layout->luc_is_rootwindow = 0;

   /* hard coded part */
    layout->DrawUpDirection = DrawUpHorizontally;
}

void
s_lookup_draw( ic, rv )
    FXimp_Client	*ic ;
    iml_inst		*rv ;
{
    XIMLookupDrawCallbackStruct	*lds =
			(XIMLookupDrawCallbackStruct *)&rv->operand ;
    LookupData *lookup_data = (LookupData *)ic->lookup_data;
    htt_t       *htt = IMS_OF_CLIENT(ic);
    Bool	ims_window_mapped = GET_HTT_VALUE(htt, ims_window_mapped);
    Display	*display = GET_HTT_VALUE(htt, display);
    LookupInfo	*lookup_info = (LookupInfo  *)GET_HTT_VALUE(htt, lookup_info);
    int		prev_count;
    int		this_count;

    if ( !lookup_data )
		 return ;

    if (!ims_window_mapped && lookup_info->style == LOOKUP_MENU_IMSWINDOW)
	MapIMSLuc(display, htt);

    prev_count = lookup_data->current_count;
    Lkup_cancel_item( ic );

    lookup_data->first_candidate = lds->index_of_first_candidate;
    lookup_data->last_candidate = lds->index_of_last_candidate;
    lookup_data->current_candidate = lds->index_of_first_candidate;
    lookup_data->n_choices = lds->n_choices;
    Lkup_make_item( ic, lds ) ;
    this_count = Lkup_get_item( ic ) ;

    if(lookup_data->lucstate == LUCStart)
	Lkup_resize_window(ic, True);
    else {
	if(this_count != prev_count)
        	Lkup_resize_window(ic, False);
	else
		Lkup_Redisplay((XEvent *)NULL, ic);
    }
    ic->select_kanji_mod = 1;
    current_lookup_ic = ic;
    lookup_data->lucstate = LUCDraw;
    return ;
}

static void
Lkup_cancel_item( ic )
    FXimp_Client	*ic ;
{
    int		 i ;
    LookupData *lookup_data = (LookupData *)ic->lookup_data;
    LookupItem	*item = lookup_data->lookup_item;

    for ( i = 0 ; i < lookup_data->n_choices ; i++, item++ ) {
	free( item->label );
	free( item->string );
	if(item->feedback)
		free(item->feedback);
    }
    free( lookup_data->lookup_item );
    if(lookup_data->title)
	free(lookup_data->title);
    if(lookup_data->title_fb)
	free(lookup_data->title_fb);
    lookup_data->lookup_item = NULL;
    return ;
}

static void
Lkup_make_item( ic, lds )
    FXimp_Client		*ic ;
    XIMLookupDrawCallbackStruct	*lds ;
{
    htt_t	*htt = IMS_OF_CLIENT(ic);
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    int		 item_count;
    int		 max_width;
    int		 candidate = lds->index_of_first_candidate;
    wchar_t	*value;
    LookupItem	*item;
    LookupData	*lookup = (LookupData *)ic->lookup_data;
    XIMChoiceObject	*choice = lds->choices;
    XFontSet	 fontset;
#ifdef  DEBUG
    dbg_printf( "<<< Lookup make_item Start: choice_ptr=%x\n", choice );
    dbg_printf( "    n_choices=%d\n", lds->n_choices );
#endif  /* DEBUG */
    fontset = Lkup_get_font( CURRENT_CLIENT(htt) );
    lookup->lookup_item = item =
		(LookupItem *)malloc( sizeof( LookupItem ) * lds->n_choices );
    bzero( item, sizeof( LookupItem ) * lds->n_choices );
    max_width = 0;

    if(lds->title) {
	lookup->title = 
		(wchar_t *)malloc(sizeof(wchar_t)*(lds->title->length + 1));
	if(lds->title->encoding_is_wchar)
		wscpy(lookup->title, lds->title->string.wide_char);
	else
		mbstowcs(lookup->title, lds->title->string.multi_byte,
				lds->title->length);
	if(lds->title->feedback) {
		lookup->title_fb = 
		(XIMFeedback *)malloc(sizeof(XIMFeedback)*(lds->title->length));
		bcopy(lds->title->feedback, lookup->title_fb,
			sizeof(XIMFeedback)*(lds->title->length));
	}
	else
		lookup->title_fb = (XIMFeedback *)NULL;
	lookup->title_len = lds->title->length;
    }
    else
	lookup->title = (wchar_t *)NULL;

    for ( item_count = 0; item_count < lds->n_choices;
			item_count++, item++, choice++, candidate++ ) {
#ifdef DEBUG
    dbg_printf( "    choice->value=%x ", choice->value );
    dbg_printf( "    choice->value->length=%d ", choice->value->length );
    dbg_printf( "    choice->value->string=%x\n", choice->value->string.wide_char );
#endif					/* DEBUG */
	item->candidate = candidate;
	if ( lookup->layout.WhoOwnsLabel == XIMOwnsLabel ) {
#ifdef DEBUG
    dbg_printf( "    choice->label=%x ", choice->label );
    dbg_printf( "    choice->label->string=%x ", choice->label->string.wide_char );
#endif					/* DEBUG */
	    item->label_length = choice->label->length;
	    item->label = (wchar_t *)malloc( sizeof( wchar_t )
					* ( choice->label->length + 1 ) );
	    if ( choice->label->encoding_is_wchar ) {
		wscpy( item->label, choice->label->string.wide_char );
	    } else {
		mbstowcs( item->label, choice->label->string.multi_byte,
					choice->label->length );
	    }
	} else {
	    item->label_length = 1;
	    item->label = (wchar_t *)malloc( sizeof( wchar_t ) * 2 );
	    item->label[0] = constant_alph[0];
	    item->label[1] = 0;
	}

	item->value_length = choice->value->length;
	if(choice->value->feedback) {
		XIMFeedback	*tmp;
		item->feedback = (XIMFeedback *)malloc( sizeof(XIMFeedback)
			 * ( item->label_length + item->value_length + 1 ) );
		tmp = item->feedback + 1 + item->label_length;
#ifdef	SVR4
		memset(item->feedback, 0, 
				(1 + item->label_length) * sizeof(XIMFeedback));
		memmove(tmp, choice->value->feedback,
				item->value_length * sizeof(XIMFeedback));
#else
		bzero(item->feedback, 
				(1 + item->label_length) * sizeof(XIMFeedback));
		bcopy(choice->value->feedback, tmp, 
				item->value_length * sizeof(XIMFeedback));
#endif
	}	
	item->string = (wchar_t *)malloc( sizeof( wchar_t )
			 * ( item->label_length + item->value_length + 2 ) );
	wscpy( item->string, item->label );
	value = item->string + item->label_length;
	value[0] = constant_alph[0];
	value++;
	if ( choice->value->encoding_is_wchar ) {
	    wscpy( value, choice->value->string.wide_char );
	} else {
	    mbstowcs( value, choice->value->string.multi_byte,
					choice->value->length );
	}
	item->string_length = item->label_length + item->value_length + 1;
	item->string_width = XwcTextEscapement(lookup_info->fontset,
					item->string, item->string_length);
	max_width = MAX( max_width, item->string_width );
    }

    lookup->next_page_width =
	XwcTextEscapement(lookup_info->fontset, lookup_info->constant_next,
					LKUP_NEXT_LEN);
    lookup->prior_page_width =
	XwcTextEscapement(lookup_info->fontset, lookup_info->constant_prior,
					LKUP_PRIOR_LEN);
    item = lookup->lookup_item;
    for ( item_count = 0; item_count < lds->n_choices;
				item_count++, item++, choice++ )
    item->string_width = max_width;
    lookup->lookup_item->page_start_flag = True;
    return;
}

static int
Lkup_get_item( ic )
    FXimp_Client	*ic;
{
    htt_t		*htt = IMS_OF_CLIENT(ic);
    LookupInfo		*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display		*display = GET_HTT_VALUE(htt, display);
    int			 item_count;
    int			 count;
    int			 nrows;
    int			 loc;
    int			 x1, y1, x2, y2;
    int			 half_font_height;
    int			 width;
    int			 lookup_width,
			 lookup_height;
    int			 max_width,
			 max_loc;
    LookupItem		*item;
    LookupData		*lookup = (LookupData *)ic->lookup_data;
    XFontSet		fontset;
    int			max_candidates_by_label_type;
    int			title_w;
    int			title_offx = 0;
    int			title_offy = 0;

    fontset = Lkup_get_font( CURRENT_CLIENT(htt) );
    half_font_height = lookup_info->font_ext->max_logical_extent.height / 2;
    if ( half_font_height < 8 )
	half_font_height = 8;
    lookup->current_count = 0;
    max_loc = loc = nrows = count = 0;
    item_count = lookup->current_candidate - lookup->first_candidate;
    item = lookup->lookup_item + item_count;
    x1 = y1 = 0;
    max_width = width = 0;
    lookup_width = lookup_info->linewidth * 2
		    + ( lookup_info->wspacing1 + lookup_info->wspacing2 )
		    * lookup_info->font_ext->max_logical_extent.width;

    if(lookup_info->style == LOOKUP_MENU_IMSWINDOW)
	lookup_height = half_font_height * 3;
    else
        lookup_height = lookup_info->linewidth * 2
			+ half_font_height * 2
			+ lookup_info->hspacing * half_font_height * 2;

    if(lookup->title && lookup_info->style != LOOKUP_MENU_IMSWINDOW) {
	title_w = XwcTextEscapement(fontset,  lookup->title, 
				    lookup->title_len);
	if(lookup->title_pos == LOOKUP_NEXT_TO_LEFT) {
		x1 = title_w + lookup_width;
		lookup->title_x = lookup_width / 2;
	}
	else {
		y1 += lookup_height;
		lookup->title_x = -1;
	}
	title_offx = x1;
	title_offy = y1;
	lookup->title_y = lookup_height;
    }

    for( ; item_count < (int)lookup->n_choices; item_count++, item++, count++ ){
	x2 = x1 + item->string_width + lookup_width;
	if (( (x2 > (int)lookup->max_position.width) && 
	      (lookup->who_is_master == CBIsMaster) )
			|| ( lookup->layout.ncolumns == loc ) ) {
	    max_loc = MAX( max_loc, loc );
	    nrows++;
	    loc = 0;
	    x1 = title_offx;
	    x2 = x1 + item->string_width + lookup_width;
	    y1 += lookup_height - 1;
	    max_width = MAX( max_width, width );
	    width = 0;
	}
	y2 = y1 + lookup_height;
	if ( ((y2 > (int)lookup->max_position.height) &&
		(lookup->who_is_master == CBIsMaster) )
			|| ( lookup->layout.nrows == nrows ) )
		break;
	item->position.x = x1;
	item->position.y = y1;
	item->position.width = x2 - x1 - 1;
	item->position.height = y2 - y1 - 1;
	item->ncolumns = loc;
	item->nrows = nrows;
	lookup->nrows = nrows;
	if ( lookup->layout.WhoOwnsLabel == CBOwnsLabel ) {
	    item->label[0] = constant_alph[count+1];
	    item->string[0] = constant_alph[count+1];
	}

	loc++;
	x1 = x2 - 1;
	width += item->position.width;
    }

    if(lookup->layout.label_type == IM_LUC_LABEL_NUMERIC)
	max_candidates_by_label_type = 10;
    else
	max_candidates_by_label_type = 26;
    if( count > max_candidates_by_label_type ) {
	count = max_candidates_by_label_type;
	item_count = lookup->current_candidate + count;
    }

    max_loc = MAX( max_loc, loc );
    lookup->ncolumns = max_loc;
    lookup->nrows = ( count - 1 )/max_loc + 1;
   
    lookup->prior_page = False;
    lookup->next_page = False;
    if ( lookup->current_candidate != lookup->first_candidate )
	lookup->prior_page = True;
    if ( item_count < lookup->n_choices ) {
	lookup->next_page = True;
    }
    lookup->current_count = count;

    if( lookup->current_count <= 0 )
	lookup->current_count = 1;

    lookup->position.width = MAX( max_width, width );
    lookup->position.width += (short)title_offx;
    lookup->position.height = lookup_height * lookup->nrows + title_offy;
    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
	lookup->position.width += LKUP_POPUP_SPACE * 2;
	lookup->position.height += LKUP_POPUP_SPACE * 2;

	if((int)lookup->position.width + (int)lookup->position.x > 
					DisplayWidth(display, ic->screen) )
		lookup->position.x = DisplayWidth(display, ic->screen) -
					lookup->position.width;
	if((int)lookup->position.height + (int)lookup->position.y >
					DisplayHeight(display, ic->screen) )
		lookup->position.y = DisplayHeight(display, ic->screen) -
					lookup->position.height;

    }

    if(lookup->title_x < 0) {
	lookup->title_x = (title_w > (int)lookup->position.width)?
		lookup_width :
		((int)lookup->position.width - title_w)/2;
    }
    lookup->lineoffset_x = title_offx;
    lookup->lineoffset_y = title_offy;

#ifdef	DEBUG
    dbg_printf( "<<< Lookup get_item2 end count=%d, ncol=%d, nrows=%d\n",
		count, lookup->ncolumns, lookup->nrows );
#endif	/* DEBUG */
    if ( lookup->ncolumns == 0 )
	return ( 1 );
    return (count);
}

static	void
Lkup_resize_window(ic, need_move)
    FXimp_Client	*ic ;
    Bool		need_move;
{
    htt_t	 *htt = IMS_OF_CLIENT(ic);
    LookupInfo	 *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display	 *display = GET_HTT_VALUE(htt, display);
    FXimp_Client *root_client = GET_HTT_VALUE(htt, root_client);
    LookupData	 *lookup = (LookupData *)ic->lookup_data;
    Bool	  need_fake = False;
    static XWindowAttributes	attr;
    static Bool	  need_frame_width = True;

#ifdef	DEBUG
    dbg_printf( "<<< Lookup resize_window Start\n" );
    dbg_printf( "    x=%d, y=%d, wid=%d, hei=%d\n",
		lookup->position.x, lookup->position.y,
		lookup->position.width, lookup->position.height );
#endif	/* DEBUG */

    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
	if(lookup_info->move)
		need_fake = True;
	if(need_fake)
		XMapWindow( display, ic->lookup_window );
	if(lookup_info->move) {
	    if(need_frame_width) {
		XGetWindowAttributes(display, ic->lookup_window, &attr);
		need_frame_width = False;
	    }
	    lookup->position.x -= attr.x;
	    lookup->position.y -= attr.y;
	}
	if(need_move) {
		XMoveResizeWindow( display, ic->lookup_window,
			lookup->position.x, lookup->position.y,
			lookup->position.width, lookup->position.height );
	}
	else
		XResizeWindow(display, ic->lookup_window, 
			lookup->position.width, lookup->position.height);
	XMapRaised(display, ic->lookup_window);
    } else {
	XSetWindowBackground( display, ic->lookup_window,
			      ic->preedit->Background );
        if (GET_HTT_VALUE(htt, lookup_takeover))
		XMoveResizeWindow( display, ic->lookup_window,
		  0, 0,
		  root_client->ims_area->width, root_client->ims_area->height );
        else
                XMoveResizeWindow( display, ic->lookup_window,
		  root_client->preedit->Area.x,
		  root_client->preedit->Area.y,
		  root_client->preedit->Area.width,
		  (GET_HTT_VALUE(htt, ims_preedit_line) >= GET_HTT_VALUE(htt, ims_line)) ?
		  root_client->ims_area->height - ((LookupInfo *)GET_HTT_VALUE(htt, lookup_info))->font_ext->max_logical_extent.height :
		  root_client->preedit->Area.height );
    }
    if(ic->lookup_window != ims_map_luc && ims_map_luc != -1) {
	if(lookup_info->style == LOOKUP_MENU_POPUPWINDOW)
        	XUnmapWindow( display, ims_map_luc);
    }
    ims_map_luc = ic->lookup_window ;

    if(!need_fake)
    	XMapRaised( display, ic->lookup_window );

    lookup->draw_flag = False;
    Lkup_Redisplay((XEvent *)NULL, ic);
    return ;
}

extern void
s_lookup_redisplay(evp, ic)
    XEvent		*evp;
    FXimp_Client	*ic ;
{
    htt_t		*htt = IMS_OF_CLIENT(ic);
    Display		*display = GET_HTT_VALUE(htt, display);
    LookupData		*lookup ;
    FXimp_Client 	*ximp_client;
    int			num;
    Bool		found = False;
    LookupInfo    *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    int			base = lookup_info->base_index;
 
    if(evp == (XEvent *)NULL) 
	ximp_client = CURRENT_CLIENT(htt);
    else {
	for (num=0; num<lookup_info->num_win ; num++) {
		if(evp->xany.window == lookup_info->luc_win[num+base]) {
			ximp_client = current_lookup_ic;
			found = True;
			break;
		}
	}
	if(!found)
		return;
    }
    if( !ximp_client || !ximp_client->lookup_data )
		return;

    lookup = (LookupData *)ximp_client->lookup_data;

    XClearWindow( display, ximp_client->lookup_window ) ;
    if(lookup->title && lookup_info->style != LOOKUP_MENU_IMSWINDOW)
    	Lkup_draw_title(ximp_client);
    Lkup_draw_line(ximp_client);
    Lkup_draw_item(ximp_client);
    lookup->draw_flag = True;
    return;
}

extern void
s_lookup_reconfigure(ic, evp)
    FXimp_Client        *ic ;
    XConfigureEvent	*evp;
{
    LookupData          *lookup = (LookupData *)ic->lookup_data;

    if(!lookup)
	return;

    lookup->position.x = evp->x;
    lookup->position.y = evp->y;
}


Lkup_draw_title(ximp_client)
    FXimp_Client *ximp_client;
{
    htt_t         *htt = IMS_OF_CLIENT(ximp_client);
    Display       *display = GET_HTT_VALUE(htt, display);
    LookupInfo    *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    LookupData    *lookup = (LookupData *)ximp_client->lookup_data;
    XFontSet       fontset;
    GC		   gc, rgc;
    int		   base = lookup_info->base_index;

    fontset = Lkup_get_font(ximp_client);
    if(lookup_info->style == LOOKUP_MENU_POPUPWINDOW) {
    	gc = lookup_info->gc[ximp_client->screen + base];
    	rgc = lookup_info->rgc[ximp_client->screen + base];
    }
    else {
	gc = lookup_info->gc[base];
	rgc = lookup_info->gc[base];
    }
    LUCXwcDrawString( display, ximp_client->lookup_window,
                            fontset, gc, rgc,
			    lookup->title_x, lookup->title_y,
                            lookup->title, lookup->title_fb, lookup->title_len);
}

static	void
Lkup_draw_line(ximp_client)
    FXimp_Client *ximp_client;
{
    htt_t	*htt = IMS_OF_CLIENT(ximp_client);
    Display	*display = GET_HTT_VALUE(htt, display);
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    int		 i;
    int		 x1, y1, x2, y2;
    int		 width, height;
    LookupData	*lookup = (LookupData *)ximp_client->lookup_data;
    LookupItem	*items;
    int          base = lookup_info->base_index;

    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
	items = lookup->lookup_item;
	x1 = LKUP_POPUP_SPACE + lookup->lineoffset_x;
	y1 = LKUP_POPUP_SPACE + lookup->lineoffset_y;
	height = items->position.height * lookup->nrows;
	y2 = y1 + height;
	for ( i = 0; i <= lookup->ncolumns; i++, x1 += items->position.width ) {
            XDrawLine( display, ximp_client->lookup_window,
                    lookup_info->gc[ximp_client->screen + base], x1, y1, x1, y2 );
	}
	x1 = LKUP_POPUP_SPACE + lookup->lineoffset_x;
	y1 = LKUP_POPUP_SPACE + lookup->lineoffset_y;
	width = items->position.width * lookup->ncolumns;
	x2 = x1 + width;
	for ( i = 0; i <= lookup->nrows ; i++, y1 += items->position.height ) {
          XDrawLine( display, ximp_client->lookup_window,
                lookup_info->gc[ximp_client->screen + base], x1, y1, x2, y1);
	}
    }
    return ;
}

static	void
Lkup_draw_item(ximp_client)
    FXimp_Client *ximp_client;
{
    htt_t	*htt = IMS_OF_CLIENT(ximp_client);
    Display	*display = GET_HTT_VALUE(htt, display);
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    int		 ncolumns, nrows;
    int		 add_x, add_y;
    int		 candidate, count;
    int		 half_font_height;
    int		 x, y;
    LookupItem	*items;
    LookupData	*lookup = (LookupData *)ximp_client->lookup_data;
    XFontSet	fontset;
    int		max_fw, max_fh;
    int		off_x, off_y;
    GC		gc, rgc;
    int		change_gc = 0;
    unsigned long	valuemask = 0;
    XGCValues	values;
    unsigned long	f_sav, b_sav;
    int		base = lookup_info->base_index;

    candidate = lookup->current_candidate - lookup->first_candidate;
    fontset = Lkup_get_font(ximp_client);

    max_fw = lookup_info->font_ext->max_logical_extent.width;
    max_fh = lookup_info->font_ext->max_logical_extent.height;
    half_font_height = max_fh/2;
    
    if ( half_font_height < 8 )
	half_font_height = 8;
    items = lookup->lookup_item + candidate;
    add_x = add_y = 0;
    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
		add_x = add_y = LKUP_POPUP_SPACE;
		gc = lookup_info->gc[ximp_client->screen + base];
		rgc = lookup_info->rgc[ximp_client->screen + base];
    }
    else {
        add_x = half_font_height;
	/**** Minus 2 from below line is a temporary one. it'llbe fixed.
	 **** 2/23/93 is@corp
	 ****/
        add_y = (GET_HTT_VALUE(htt, lookup_takeover)) ? half_font_height - 2 : 0;
	gc = lookup_info->gc[base];
	rgc = lookup_info->rgc[base];
	if( ximp_client->style & ( XIMPreeditCallbacks | XIMPreeditNone ) )
		;
	else {
		valuemask = GCForeground | GCBackground;
		XGetGCValues(display, gc, valuemask, &values);
		f_sav = values.foreground;
		b_sav = values.background;
		values.foreground = ximp_client->preedit->Foreground;
		values.background = ximp_client->preedit->Background;
		XChangeGC(display, gc, valuemask, &values);
	}
    }

    off_x = max_fw * lookup_info->wspacing1 + add_x;
    off_y = lookup_info->hspacing * half_font_height
	    - lookup_info->font_ext->max_logical_extent.y
	    + add_y;

    count = 1;
    for ( nrows = 0; nrows <= lookup->nrows; nrows++ ) {
	for ( ncolumns = 0; ncolumns <= lookup->ncolumns; ncolumns++ ) {
	    if ( count > lookup->current_count )
			 break;
	    x = items->position.x + off_x;
	    y = items->position.y + off_y; 
            LUCXwcDrawString( display, ximp_client->lookup_window,
                            fontset, gc, rgc,
                            x, y, items->string, items->feedback,
			    items->string_length );
	    items++; count++;
	}
    }
    if(valuemask) {
	values.foreground = f_sav;
	values.background = b_sav;
	XChangeGC(display, gc, valuemask, &values);
    }
    XFlush( display ) ;
    return ;
}

LUCXwcDrawString(display, win, fontset, gc, rgc, x, y, str, fb, len)
    Display	*display;
    Window	win;
    XFontSet	fontset;
    GC		gc;
    GC		rgc;
    int		x;
    int		y;
    wchar_t	*str;
    XIMFeedback	*fb;
    int		len;
{
    XIMFeedback	*f1 = fb;
    XIMFeedback	*f2;
    int		relen = len;
    int		fg_len;
    int		width;

    if(!fb) {
	XwcDrawString(display, win, fontset, gc, x, y, str, len);
	return;
    }

    while(relen) {
	f2 = f1;
	while(*f2 == *f1 && relen) {
		f2++;
		relen--;
	}
	fg_len = len - relen;
	if(*f1 & XIMReverse)
	    XwcDrawImageString(display, win, fontset, rgc, x, y, str, fg_len);
	else
	    XwcDrawString(display, win, fontset, gc, x, y, str, fg_len);
    	width = XwcTextEscapement(fontset, str, fg_len);
	if(*f1 & XIMUnderline)
		XDrawLine(display, win, gc, x, y, x + width, y);
	str += fg_len;
	x += width;
	f1 = f2;
	len = relen;
    }
}

void
s_lookup_process( ic, rv )
    FXimp_Client	*ic ;
    iml_inst		*rv ;
{
    htt_t		*htt = IMS_OF_CLIENT(ic);
    int			 i ;
    int			 candidate;
    char		 lookup_buff[4] ;
    wchar_t		 select_label[4] ;
    XKeyEvent		*e ;
    KeySym		 key ;
    XComposeStatus	 cs ;

    XIMLookupProcessCallbackStruct	*lookup_process = 
			(XIMLookupProcessCallbackStruct *)&rv->operand;
    LookupData		*lookup = (LookupData *)CURRENT_CLIENT(htt)->lookup_data;
    int			 current_num = lookup->current_count;
    LookupItem		*current_item;
    unsigned int	save_state;

#ifdef	DEBUG
    dbg_printf( "<<< Lookup process Start\n" );
#endif
    if ( !ic->lookup_data ) return ;

    lookup->lucstate = LUCProc;
    candidate = lookup->current_candidate - lookup->first_candidate;
    current_item = lookup->lookup_item + candidate;
    e = (XKeyEvent *)lookup_process->event ;
    if ( lookup->draw_flag == False ) {
	lookup_process->index_of_choice_selected = XIM_UNKNOWN_KEYSYM ;
	return ;
    }
    bzero( lookup_buff, sizeof( lookup_buff ) ) ;
    bzero( select_label, sizeof( select_label ) ) ;

    save_state = e->state;
    e->state &= ~e->display->mode_switch;
    XLookupString(e, lookup_buff, sizeof(lookup_buff), &key, (XComposeStatus *)NULL);
    e->state = save_state;

    if ( e->state & ControlMask ) {
	if ( ( key == XK_P || key == XK_p )
	  && ( lookup->prior_page == True ) ) {
	    current_item--;
	    for ( ; !current_item->page_start_flag; current_item-- );
	    lookup->current_candidate = current_item->candidate;
	    if( current_num != Lkup_get_item( ic ))
	    	Lkup_resize_window(ic, False) ;
	    Lkup_Redisplay((XEvent *)NULL, ic);
	    lookup_process->index_of_choice_selected = XIM_UNDETERMINED ;
	} else if ( ( key == XK_N || key == XK_n )
		 && ( lookup->next_page == True ) ) {
	    lookup->current_candidate += lookup->current_count;
	    current_item += lookup->current_count;
	    current_item->page_start_flag = True;
	    if( current_num != Lkup_get_item( ic ))
	    	Lkup_resize_window(ic, False) ;
	    Lkup_Redisplay((XEvent *)NULL, ic);
	    lookup_process->index_of_choice_selected = XIM_UNDETERMINED ;
	} else {
	    lookup_process->index_of_choice_selected = XIM_UNKNOWN_CONTROL;
	}
    } else {
	mbstowcs( select_label, lookup_buff, 1 ) ;
	for ( i = 0; i < lookup->current_count; i++, current_item++ ) {
	    if ( xsf_wcscmp( select_label, current_item->label ) == 0 ) {
		lookup_process->index_of_choice_selected =
						current_item->candidate ;
		Lkup_End( ic ) ;
		return;
	    }
	}
	if ( i == lookup->current_count )
	    lookup_process->index_of_choice_selected = XIM_UNKNOWN_KEYSYM ;
	return ;
    }
}

void
s_lookup_done( ic )
    FXimp_Client	*ic ;
{
    htt_t	*htt = IMS_OF_CLIENT(ic);
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);

    lookup_info->lock --;
    if ( !ic->lookup_data ) return ;
    Lkup_End( ic ) ;
    return ;
}

extern void
s_lookup_focusin( ic )
    FXimp_Client	*ic;
{
    htt_t	*htt = IMS_OF_CLIENT(ic);
    Display	*display = GET_HTT_VALUE(htt, display);
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    FXimp_Client	*root_client = GET_HTT_VALUE(htt, root_client);
    Bool	ims_window_mapped = GET_HTT_VALUE(htt, ims_window_mapped);
    LookupData	*lookup = (LookupData *)CURRENT_CLIENT(htt)->lookup_data;

#ifdef	DEBUG
    dbg_printf( "<<< Lookup focusin Start mod=%d\n",
		ic->select_kanji_mod );
#endif	/* DEBUG */
    if (!ims_window_mapped && lookup_info->style == LOOKUP_MENU_IMSWINDOW)
	MapIMSLuc(display, htt);

    if ( ( ic->lookup_data ) && ( ic->select_kanji_mod == 1 ) ) {
	if ( lookup_info->style == LOOKUP_MENU_IMSWINDOW ) {
	    if ((GET_HTT_VALUE(htt, preedit_feedback) != FEEDBACK_NONE)
		&& (!(ic->style & XIMPreeditCallbacks)))
		stipple_or_unmap_preedit_window(ic, FEEDBACK_UNMAP);
	    if ((GET_HTT_VALUE(htt, status_feedback) != FEEDBACK_NONE)
		&& (!(ic->style & XIMStatusCallbacks))
		&& (GET_HTT_VALUE(htt, lookup_takeover)))
		stipple_or_unmap_status_window(ic, FEEDBACK_UNMAP);
	    lookup->max_position.width = (GET_HTT_VALUE(htt, lookup_takeover)) ? root_client->ims_area->width : root_client->preedit->Area.width;
	    if ( ic->style & ( XIMPreeditCallbacks | XIMPreeditNone ) ) {
		XSetWindowBackground( display, ic->lookup_window,
		root_client->preedit->Background );
	    } else {
		XSetWindowBackground( display, ic->lookup_window,
		    ic->preedit->Background );
	    }
	    /****
	    *XClearArea( display, ic->lookup_window, 0, 0, 0, 0, True );
	    ****/
	}
	Lkup_get_item( ic );
	Lkup_resize_window(ic, False /**** True ****/);
	current_lookup_ic = ic;
    }

    return;
}

static void
MapIMSLuc(display, htt)
    Display	*display;
    htt_t	*htt;
{
    FXimp_Client	*root_client = GET_HTT_VALUE(htt, root_client);

    XMapRaised(display, root_client->window);
    SET_HTT_VALUE(htt, ims_window_mapped, True);
    ims_map_luc = root_client->window;
}

extern void
s_lookup_focusout( ic )
    FXimp_Client	*ic;
{
    htt_t		*htt = IMS_OF_CLIENT(ic);
    Display		*display = GET_HTT_VALUE(htt, display);
    Bool 		ims_window_mapped = GET_HTT_VALUE(htt, ims_window_mapped);
    FXimp_Client	*root_client = GET_HTT_VALUE(htt, root_client);
    LookupInfo          *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);

#ifdef	DEBUG
    dbg_printf( "<<< Lookup focusout Start mod=%d\n",
			ic->select_kanji_mod );
#endif	/* DEBUG */

    if ( ( ic->lookup_data )
      && ( ic->select_kanji_mod == 1 )
      && ( lookup_info->style == LOOKUP_MENU_IMSWINDOW ) )
	XUnmapWindow( display, ic->lookup_window ) ;
    if ( lookup_info->style == LOOKUP_MENU_IMSWINDOW && !ims_window_mapped ) {
	XMapRaised( display, root_client->window );
	SET_HTT_VALUE(htt, ims_window_mapped, True);
    }

    return;
}

void
s_lookup_end( ic )
    FXimp_Client	*ic ;
{
    htt_t	 *htt = IMS_OF_CLIENT(ic);
    Display	 *display = GET_HTT_VALUE(htt, display);
    Bool 	 ims_window_mapped = GET_HTT_VALUE(htt, ims_window_mapped);
    FXimp_Client *root_client = GET_HTT_VALUE(htt, root_client);
    LookupInfo	 *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);

    if (!ic->lookup_data) 
		return ;
    Lkup_cancel_item( ic ) ;
    XMoveResizeWindow( display, ic->lookup_window,-100,-100,1,1);
    XFlush(display);
    free(ic->lookup_data);
    ic->lookup_data = 0 ;
    ic->select_kanji_mod = 0;
    if ( !ims_window_mapped && lookup_info->style == LOOKUP_MENU_IMSWINDOW) {
	XMapRaised( display, root_client->window );
	ims_window_mapped = SET_HTT_VALUE(htt, ims_window_mapped, True);
        ims_map_luc = root_client->window;
    }
    if ((!(ic->style & XIMStatusCallbacks))
        && (GET_HTT_VALUE(htt, lookup_takeover)))
    {
	MapStatusWindow(ic);
	draw_status_area(ic, highlight);
    }
    return ;
}

void
s_lookup_destroy( ic )
    FXimp_Client	*ic ;
{
    if ( !ic->lookup_data ) return ;
    Lkup_End( ic ) ;
    return ;
}

static int
xsf_wcscmp( wcs1, wcs2 )
    wchar_t	*wcs1;
    wchar_t	*wcs2;
{
#ifdef	DEBUG
    dbg_printf( "xsf_wcscmp wcs1=%8.8x, wcs2=%8.8x, wcs1=%s, wcs2=%s\n",
			wcs1, wcs2, wcs1, wcs2 );
#endif	/* DEBUG */
    for ( ; ( *wcs1 != 0 ) || ( *wcs2 != 0 ) ; wcs1++, wcs2++ ) {
	if ( *wcs1 != *wcs2 )
	    return ( 1 );
    }
    if ( ( *wcs1 == 0 ) && ( *wcs2 == 0 ) )
	return ( 0 );
    return ( 1 );
}

static XFontSet
Lkup_get_font( current_client )
    FXimp_Client	 *current_client;
{
    htt_t	  *htt = IMS_OF_CLIENT(current_client);
    LookupInfo	  *lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    FXimp_Client  *root_client = GET_HTT_VALUE(htt, root_client);
    XFontSet	   fontset;

    if ( lookup_info->style == LOOKUP_MENU_POPUPWINDOW ) {
	fontset = lookup_info->fontset;
    } else if ( current_client->style & ( XIMPreeditCallbacks | XIMPreeditNone ) ) {
	fontset = root_client->preedit_font->font_set;
    } else {
        fontset = lookup_info->fontset;
    }

    return fontset;
}

Bool
s_lookup_keyevent(ev, htt)
    XKeyEvent	*ev;
    htt_t	*htt;
{
    int		i;
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);
    Display	*display = GET_HTT_VALUE(htt, display);
    Window	w;
    long	mask;
    Window	ww;
    int		rvt;
    int		base = lookup_info->base_index;

    if(!current_lookup_ic) {
	return True;
    }

    w = current_lookup_ic->focus_window;
    for(i=0; i<lookup_info->num_win; i++) {
	if(ev->window == lookup_info->luc_win[i+base] ) {
		XGetInputFocus(display, &ww, &rvt);
		XSetInputFocus(display, w, RevertToNone, CurrentTime);
    		SET_HTT_VALUE(htt, current_client, current_lookup_ic);
    		SET_HTT_VALUE(htt, focus_in_client, current_lookup_ic);
		ev->window = w;
		if(ev->type == KeyRelease) {
			XSendEvent(display, w, True, KeyReleaseMask, ev);
		}
		else if(ev->type == KeyPress) {
			XSendEvent(display, w, True, KeyPressMask, ev);
		}
		XSetInputFocus(display, ww, rvt, CurrentTime);
		XFlush(display);
		return True;
	}
    }
    return False;
}

s_lookup_getval(htt, attr, val)
    htt_t	*htt;
    char	*attr;
    void	*val;
{
    if(!strcmp(attr, LUC_LAYOUT)) {
	GetLayout(htt, (LayoutInfo *)val);
	return	1;
    }
    else
	return 0;
}

s_lookup_setval(htt, attr, val)
    htt_t	*htt;
    char	*attr;
    void	*val;
{
    LookupInfo	*lookup_info = (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);

    if(!strcmp(attr, LUC_LOOKUP_INFO)) {
	if(lookup_info->style == LOOKUP_MENU_POPUPWINDOW)
		SwitchToPopupStyle(htt, lookup_info);
	else
		SwitchToRootStyle(htt, lookup_info);
    	htt->biff &= ~(LUCSTYLE | LUCNROWS | LUCNCOLUMNS);
	return 1;
    }
    else
	return 0;
}
